CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     Compute a 3D real-space correlation function 
C     of objects in the input file.
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      program cf

      implicit none

      include 'cf.h'
C===========================

      call cf_driver()
      
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     driver for cf
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine cf_driver()

      implicit none

      include 'cf.h'

      character*256 infile
      character*256 outfile

      integer outunit, ioflag
      parameter( outunit = 90 )

      character*16 chtemp

      real*4 rlow, rhigh

      integer nbins

      integer iargc
      external iargc

C======================

C----------------------
C     check arguments, if ok read in args.
C----------------------
      if (iargc().ne.5) then
         print*,' '
         print*,' USAGE: inputfile outputfile rlow rhigh nbins '
         print*,' '
         STOP
      endif

      call getarg(1,infile)
      call getarg(2,outfile)

      call getarg(3,chtemp)
      read(chtemp,*) rlow
      
      call getarg(4,chtemp)
      read(chtemp,*) rhigh

      call getarg(5,chtemp)
      read(chtemp,*) nbins

C----------------------
C     read in particle data.
C----------------------

      call get_xyz( infile )

C----------------------
C     get xi(r)
C----------------------

      call openoutput( outfile , outunit )

      call get_cf( rlow, rhigh, nbins, outunit )
      
      close( outunit )

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     get correlation function, 
C     computes weighted and unweighted 
C     1+xi(r) and errors.
C
C     assumes particle data already read into memory.
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine get_cf( rlow, rhigh, Nbins, ounit )

      implicit none

      real*4 pi4on3
      parameter ( pi4on3 = 4.188790205D0 )

      integer Nbins, lcv, lcvoct, ounit, lcv1, lcv2

      real*4 rinbin( Nbins )
      real*4 onepxi( 9, Nbins ), erronepxi( Nbins )
      real*4 onepxiw( 9, Nbins ), erronepxiw( Nbins )
      real*4 xidum, xiwdum, errxidum, errxiwdum

      real*4 rlow, rhigh
      real*4 lgrlow, lgrhigh, dlgr

      real*4 rleft, rright, dV

      real*4 rand, wrand

      real*4 rt, wrt, poisseb, npreal

      include 'cf.h'

C====================================================

      npreal = real( nptcl )

      poisseb = sqrt( npreal )

      rand = ( npreal * ( npreal - 1) / 2. )/( LBOXH**3 )

      wrand = wp2mean * 
     & ( npreal * (npreal - 1 ) / 2. )/( LBOXH**3 ) 

      lgrlow = log10(rlow)
      lgrhigh = log10(rhigh)
      dlgr = (lgrhigh - lgrlow) / real( Nbins )

C-----------------
C     count all pairs
C-----------------

      call count_pairs( rlow, rhigh, Nbins, rinbin )

C-----------------
C     xi(r) for box
C-----------------

      do lcv = 1, Nbins

         rright = lgrlow + real(lcv)*dlgr
         rleft = rright - dlgr
         rright = 10.0**rright
         rleft = 10.0**rleft
         
         dV = pi4on3*( rright**3 - rleft**3 )

         rt = rand*dV
         wrt = wrand*dV

         xidum = npairs(1,lcv)/rt - 1.
         xiwdum = nwpairs(1,lcv)/wrt - 1.

         poisseb = sqrt( npairs( 1, lcv ) ) !> poisson errorbar weight <!
         
         rt = 7.*rt/8.
         wrt = 7.*wrt/8.

         errxidum = 0.
         errxiwdum = 0.

         do lcvoct = 2,9

            onepxi( lcvoct, lcv ) = 
     & npairs( lcvoct, lcv )/rt - 1.

            errxidum = errxidum + 
     & ( onepxi( lcvoct , lcv ) - xidum )**2

            onepxiw( lcvoct, lcv ) = 
     &  nwpairs( lcvoct, lcv )/wrt - 1.

            errxiwdum = errxiwdum + 
     & ( onepxiw( lcvoct, lcv ) - xiwdum)**2

         enddo

         if (npairs(1,lcv).gt.0.) then
            rinbin(lcv) = rinbin(lcv)/npairs(1,lcv)
         else
            rinbin(lcv) = 0.
         endif

C------------------------
C     Set error = jackknife error plus poisson error
C------------------------

         errxidum = sqrt( 8.*errxidum/7. ) + (1. + xidum)/poisseb

         errxiwdum = sqrt( 8.*errxiwdum/7. ) + (1. + xiwdum)/poisseb

         onepxi(1, lcv) = xidum !+ 1. output xi(r) rather than 1+xi(r)
         erronepxi(lcv) = errxidum

         onepxiw(1, lcv) = xiwdum !+ 1. output xi(r)
         erronepxiw(lcv) = errxiwdum

      enddo

C-------------------------
C     Calculation complete now output.
C-------------------------

      do lcv=1,Nbins
         
         if (rinbin(lcv).gt.0.) then

            write( ounit, 901 ) 
     & rinbin(lcv), onepxi( 1, lcv ), erronepxi( lcv ), 
     & (onepxi( lcv1 , lcv ), lcv1=2,9 ), onepxiw( 1, lcv), 
     & erronepxiw( lcv ) 

         endif

      enddo

 901  format(1X,F9.4,2X,10(E10.4,2X),2(E10.4,2X))

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     count pairs
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine count_pairs( rlow, rhigh, Nbins, rinbin )

      implicit none

      include 'cf.h'

      integer lcvp1, lcvp2, lcvoct

      integer Nbins, binnumber, poct 

      real*4 rinbin( Nbins )

      real*4 rlow, rhigh, separation, lgsep
      real*4 lgRlow, lgRhigh, dlgR

      real*4 dist2pts
      external dist2pts
C========================

C-----------------------
C      initialize
C-----------------------

      lgRlow = log10(rlow)
      lgRhigh = log10(rhigh)

      dlgR = ( lgRhigh - lgRlow ) / real( Nbins )
      
      call init_array( rinbin , Nbins )
      call init_2D_array( npairs, 9, Nmaxbin )
      call init_2D_array( nwpairs, 9, Nmaxbin )

C-----------------------
C     count pairs
C-----------------------

      do lcvp1=1,nptcl
         do lcvp2=lcvp1,nptcl

C-----------------------
C     Get separation
C-----------------------

            separation = dist2pts( 
     & xyzp(1,lcvp1), xyzp(2,lcvp1), xyzp(3,lcvp1), 
     & xyzp(1,lcvp2), xyzp(2,lcvp2), xyzp(3,lcvp2), 
     & LBOXH )

C-----------------------
C     If in range bin it up.
C-----------------------
            if ((separation.lt.rhigh)
     &                    .and.
     &          (separation.ge.rlow)) then
            
               lgsep = log10( separation )

               binnumber = max( 1, 
     & min( Nbins, (int( (lgsep - lgRlow) / dlgR ) + 1 ) ) )

               npairs( 1, binnumber ) = 
     & npairs( 1, binnumber ) + 1.

               nwpairs( 1, binnumber ) = 
     & nwpairs( 1, binnumber ) + wp(lcvp1)*wp(lcvp2)

               rinbin( binnumber ) = rinbin( binnumber ) + separation
C-----------------------
C     Add to octant pairs if neither are in the 
C     excluded octant.
C-----------------------

               do lcvoct=1,8

                  if ( (octant(lcvp1).ne.lcvoct)
     &                              .and.
     &                 (octant(lcvp2).ne.lcvoct) ) then

                     poct = 1 + lcvoct

                     npairs( poct , binnumber ) = 
     & npairs( poct , binnumber ) + 1.
                     nwpairs( poct , binnumber ) = 
     & nwpairs( poct , binnumber ) + wp(lcvp1)*wp(lcvp2)

                  endif

               enddo

            endif
         enddo
      enddo

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     initialize particle positions and octant labels.
C     build a chaining mesh for particle locations.
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine get_xyz( dfile )
      
      implicit none

      include 'cf.h'
      
      character*256 dfile
      integer dunit, ioflag, lcv, lcvdim, poct
      parameter( dunit = 10 )

      real*4 xread, yread, zread, wread, boxon2, wsum, w2sum
C==============================

C-- Initialize chaining mesh - not adaptive
      Ngrid=Ngridmax
      Nmesh=Ngrid**3

      call opendata( dfile , dunit )

      ioflag = 0
      lcv = 0
      wsum = 0.
      w2sum = 0.

      boxon2 = LBOXH / 2.

      do while (ioflag.eq.0) 

         read(dunit,*,iostat=ioflag)
     & xread, yread, zread, wread

         if (ioflag.eq.0) then
            
            lcv = lcv + 1
c... assign position
            xyzp(1,lcv) = xread
            xyzp(2,lcv) = yread
            xyzp(3,lcv) = zread
            wp(lcv) = wread

            wsum = wsum + wread
            w2sum = w2sum + wread**2

c... assign octant
            poct = 1

            do lcvdim = 1 , NDIM
               if (xyzp(lcvdim,lcv).le.boxon2) 
     &              poct = poct + 2**( lcvdim - 1 )
            enddo

            octant(lcv) = poct

         endif

      enddo

      nptcl = lcv
      wpmean = wsum / real( nptcl )
      wp2mean = w2sum / real( nptcl )

      print*,' '
      print*,'    DATA READ COMPLETE, Number of particles = ',lcv
      print*,' '

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine opendata(dfile,dunit)

      implicit none

      character*256 dfile
      integer dunit, ioflag, cpos

      logical opflag
C=========================================

      inquire(unit=dunit,opened=opflag)

      if (opflag) then
         print*,' '
         print*,' ERROR: This unit is already '
         print*,'        opened.'
         print*,'   unit = ',dunit,' '
         print*,'   '
         STOP
      endif
C...
      
      cpos = index(dfile,' ')-1

      open( unit=dunit,file=dfile(1:cpos),
     & status='old',form='formatted',
     & iostat=ioflag )
      
      if (ioflag.ne.0) then

         print*,' '
         print*,' ERROR: Could not open data file.'
         print*,' '
         print*,'      file = '//dfile(1:cpos)//' '
         print*,'      unit = ',dunit
         print*,' '
         STOP
         
      endif

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine openoutput(ofile,ounit)

      implicit none

      character*256 ofile
      integer ounit, ioflag, cpos

      logical opflag
C=============================================

      inquire(unit=ounit,opened=opflag)

      if (opflag) then
         print*,' '
         print*,' ERROR: This unit is already opened '
         print*,' '
         print*,'       unit = ',ounit
         print*,' '
         STOP
      endif
C...

      cpos = index(ofile,' ')-1

      open(unit=ounit,file=ofile(1:cpos),
     & status='unknown',form='formatted',
     & iostat=ioflag)
      
      if (ioflag.ne.0) then
         print*,' '
         print*,' ERROR: Could not open data file.'
         print*,' '
         print*,'      file = '//ofile(1:cpos)//' '
         print*,'      unit = ',ounit
         print*,' '
         STOP
      endif

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     distance between points 1 and 2 in the box.
C     box the box size in the same units 
C     as the x,y,z values.  
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

      function dist2pts(x1,y1,z1,x2,y2,z2,box)

      implicit none

      real*4 dist2pts, x1, y1, z1, x2, y2, z2, boxon2
      real*4 box

      real*4 dx, dy, dz, shx, shy, shz
C=====================================================
C
      boxon2=box/2.

      dx = x1-x2
      dy = y1-y2
      dz = z1-z2

      if (dx.ge.boxon2) then
         shx = -box
      elseif (dx.lt.-boxon2) then
         shx = box
      else
         shx = 0.
      endif
      
      if (dy.ge.boxon2) then
         shy = -box
      elseif (dy.lt.-boxon2) then
         shy = box
      else
         shy = 0.
      endif

      if (dz.ge.boxon2) then
         shz = -box
      elseif (dz.lt.-boxon2) then
         shz = box
      else
         shz = 0.
      endif

      dist2pts = sqrt( (dx + shx)**2 + 
     & (dy + shy)**2 + (dz + shz)**2 )

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine init_array( vec , nvec )
      implicit none
      
      integer nvec, lcv
      real*4 vec( nvec )

      do lcv=1,nvec
         vec(lcv) = 0.
      enddo
      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      subroutine init_2D_array( vec , n1, n2 )

      implicit none
      
      integer n1, n2, lcv1, lcv2
      real*4 vec( n1, n2 )

      do lcv1=1,n1
         do lcv2=1,n2
            vec(lcv1,lcv2) = 0.
         enddo
      enddo

      RETURN
      END
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
